$Point::nOffset = "0 0.01 0";
$Point::neOffset = "0.01 0.01 0";
$Point::eOffset = "0.01 0 0";
$Point::seOffset = "0.01 -0.01 0";
$Point::sOffset = "0 -0.01 0";
$Point::swOffset = "-0.01 -0.01 0";
$Point::wOffset = "-0.01 0 0";
$Point::nwOffset = "-0.01 0.01 0";

exec("./Support_ShapeLines.cs");
exec("./bricks.cs");
exec("./switchBricks.cs");
exec("./utils.cs");
exec("./events.cs");

$RailroadDebug = 0;



/////////
//utils//
/////////

function rotationFromAngleID(%angle)
{
	switch(%angle)
	{
		case 0:
			%rotation = "1 0 0 0";
		case 1:
			%rotation = "0 0 1 90";
		case 2:
			%rotation = "0 0 1 180";
		default:
			%rotation = "0 0 -1 90";
	}
	return %rotation;
}

function absDistanceFromVector(%start, %end, %pos) 
{
	//calculate parallelogram area of the three points and divide by base to get height
	%startToEnd = vectorSub(%end, %start);
	%startToPos = vectorSub(%pos, %start);
	%normal = vectorNormalize(%startToEnd);

	//projection of startToPos onto normal to determine location on line defined by startToEnd
	%dot = vectorDot(%startToPos, %normal);
	%projPosition = vectorAdd(vectorScale(%normal, %dot), %start);
	//behind the starting point
	if (%dot < 0) 
	{
		%dist = 1000;
	}
	else if (mAbs(%dot/vectorLen(%startToEnd)) > 1)  //past the ending point
	{
		%dist = 1000;
	}
	else
	{
		//calculate parallelogram area of the three points and divide by base to get height (eg dist from startToEnd)
		%area = vectorLen(vectorCross(%startToEnd, %startToPos));
		%dist = %area/vectorLen(%startToEnd);		
	}
	return %dist SPC %area SPC %projPosition;
}

function serverCmdTrain(%cl) 
{
	if (!%cl.isAdmin)
		return;

	if (isObject(%pl = %cl.player))
	{
		if (isObject(%mount = %pl.getObjectMount()))
		{
			trainLockLoop(%mount);
			messageClient(%cl, '', "\c5Converted " @ trim(%mount.getDatablock().uiName) @ " to a train object");
		}
		else
		{
			trainLockLoop(%pl);
			messageClient(%cl, '', "\c5Converted you to a train object");
		}
	}
	else
	{
		messageClient(%cl, '', "You need to have a player or a vehicle to use /train!");
	}
}

function serverCmdUnTrain(%cl) 
{
	if (!%cl.isAdmin)
		return;

	if (isObject(%pl = %cl.player))
	{
		if (isObject(%mount = %pl.getObjectMount()))
		{
			cancel(%mount.trainLoopSchedule);
			messageClient(%cl, '', "\c5Converted " @ %mount.getDatablock().uiName @ " to a normal object");
		}
		else
		{
			cancel(%pl.trainLoopSchedule);
			messageClient(%cl, '', "\c5Converted you to a normal object");
		}
	}
	else
	{
		messageClient(%cl, '', "You need to have a player or a vehicle to use /untrain!");
	}
}

package Railroad
{
	function fxDtsBrick::spawnVehicle(%brick, %vehicle)
	{
		%val = Parent::spawnVehicle(%brick, %vehicle);
		if (isObject(%brick.vehicle) && %brick.vehicle.getDatablock().isTrain)
		{
			trainLockLoop(%brick.vehicle);
		}
		return %val;
	}

	function fxDTSBrick::plant(%brick)
	{
		%ret = parent::plant(%brick);
		if (!%ret && %brick.getDatablock().isTrainTrack)
		{
			%brick.schedule(100, createTrack);
		}
		return %ret;
	}

	function fxDTSBrick::onAdd(%brick)
	{
		if (%brick.getDatablock().isTrainTrack)
		{
			%brick.schedule(100, createTrack);
		}
		return parent::onAdd(%brick);
	}
};
activatePackage(Railroad);

function getCompassVec(%cl) {
	%fv = vectorScale(%cl.getControlObject().getForwardVector(), 1 / mSqrt(2));
	%v = vectorAdd(%fv, "0.5 0.5 0");
	return mFloor(getWord(%v, 0)) SPC mFloor(getWord(%v, 1)) SPC "0";
}

////////
//core//
////////

function fxDTSBrick::createTrack(%brick, %name)
{
	if ($RailroadDebug)
	{
		announce("Creating track brick...");
	}

	%db = %brick.getDatablock();
	if (!%db.isTrainTrack)
	{
		return;
	}

	//save switch status
	if (%db.switchMode !$= "")
	{
		%brick.isSwitch = 1;
	}
	
	//search for adjacent track bricks

	//get directions to search
	%rot = (%brick.getAngleID() + 1) % 4;
	%pattern = %db.pattern;
	%connections = %db.connections;

	%clockwise = "N NE E SE S SW W NW N NE E SE S SW W NW";
	for (%i = 0; %i < getWordCount(%connections); %i++)
	{
		%currDir = getWord(%connections, %i);
		for (%j = 0; %j < getWordCount(%clockwise); %j++)
		{
			%search = getWord(%clockwise, %j);
			if (%search $= %currDir)
			{
				break;
			}
		}

		//rotate current direction based on angleID (90 deg)
		%j = %j + %rot * 2;

		%dynamicDir = %dynamicDir SPC getWord(%clockwise, %j);
	}
	%dynamicDir = trim(%dynamicDir);
	%dir = %dynamicDir;

	//translate directions into vectors
	%xOffset = %db.brickSizeX * 0.5 * 0.5 SPC "0 0";
	%yOffset = "0" SPC %db.brickSizeY * 0.5 * 0.5 SPC "0";
	%zOffset = "0 0" SPC %db.brickSizeZ * 0.2 * 0.5;
	%center = %brick.getPosition();

	%Pointn 	= vectorAdd(%center, %yOffset);
	%Pointne 	= vectorAdd(vectorAdd(%center, %yOffset), %xOffset);
	%Pointe 	= vectorAdd(%center, %xOffset);
	%Pointse 	= vectorAdd(vectorSub(%center, %yOffset), %xOffset);
	%Points 	= vectorSub(%center, %yOffset);
	%Pointsw 	= vectorSub(vectorSub(%center, %yOffset), %xOffset);
	%Pointw 	= vectorSub(%center, %xOffset);
	%Pointnw 	= vectorSub(vectorAdd(%center, %yOffset), %xOffset);

	//save spot where search for next rail piece will be
	%Pointn 	= %brick.Pointn 	= vectorAdd(%Pointn, $Point::nOffset);
	%Pointne 	= %brick.Pointne 	= vectorAdd(%Pointne, $Point::neOffset);
	%Pointe 	= %brick.Pointe 	= vectorAdd(%Pointe, $Point::eOffset);
	%Pointse 	= %brick.Pointse 	= vectorAdd(%Pointse, $Point::seOffset);
	%Points 	= %brick.Points 	= vectorAdd(%Points, $Point::sOffset);
	%Pointsw 	= %brick.Pointsw 	= vectorAdd(%Pointsw, $Point::swOffset);
	%Pointw 	= %brick.Pointw 	= vectorAdd(%Pointw, $Point::wOffset);
	%Pointnw 	= %brick.Pointnw 	= vectorAdd(%Pointnw, $Point::nwOffset);

	%brick.dynamicDir = %dynamicDir;

	//place debug staticshapes at each active search point
	if ($RailroadDebug)
	{
		// for (%i = 0; %i < getWordCount(%dir); %i++)
		// {
		// 	%currDir = getWord(%dir, %i);
		// 	%pos = %Point[%currDir];

		// 	%shape = createBoxAt(%pos, "1 0 0 1", 0.05);
		// 	%shape.setShapeName(getSubStr(%brick.getID(), 0, 5) SPC strUpr(%currDir));
		// }

		for (%i = 0; %i < getWordCount(%dynamicDir); %i++)
		{
			%currDir = getWord(%dynamicDir, %i);
			%pos = %Point[%currDir];

			%shape = createBoxAt(%pos, "0 1 0 1", 0.08);
			%shape.setShapeName(getSubStr(%brick.getID(), 0, 5) SPC strUpr(%currDir));
		}

		echo("Dir: " @ %dir);
		echo("DynamicDir: " @ %dynamicDir);
	}

	//search for bricks and attach if they have opposite side

	for (%i = 0; %i < getWordCount(%dir); %i++)
	{
		%currDir = getWord(%dir, %i);

		for (%j = 0; %j < getWordCount(%clockwise); %j++)
		{
			%search = getWord(%clockwise, %j);
			if (%search $= %currDir)
			{
				break;
			}
		}
		%oppositeDir = getWord(%clockwise, %j + 4);

		initContainerBoxSearch(%Point[%currDir], "0.1 0.1 0.41", $TypeMasks::fxBrickObjectType);

		while (isObject(%found = containerSearchNext()))
		{
			if (%found == %brick || !%found.getDatablock().isTrainTrack)
			{
				continue;
			}

			if (strPos(" " @ %found.dynamicDir @ " ", " " @ %oppositeDir @ " ") >= 0)
			{
				if ($RailroadDebug)
				{
					talk("\c6Track found at " @ %oppositeDir @ ", linking to " @ %currDir @ "...");
				}
				%brick.Brick[%currDir] = %found;
				%found.Brick[%oppositeDir] = %brick;
			}
		}
	}
}

function createDebugNodeMap(%brick)
{
	%stack = %brick;
	%visited = " ";
	while (%stack !$= "" && %safety < 10000)
	{
		%safety++;
		//pop off stack
		%brick = getWord(%stack, 0);
		%stack = removeWord(%stack, 0);

		//check if visited, skip if so
		if (strPos(%visited, " " @ %brick @ " ") >= 0)
		{
			continue;
		}
		%visited = %visited @ %brick @ " ";

		//create shapes
		createTrackNodeShapes(%brick);

		%dir = %brick.dynamicDir;

		for (%i = 0; %i < getWordCount(%dir); %i++)
		{
			%currDir = getWord(%dir, %i);
			%stack = trim(%stack SPC %brick.Brick[%currDir]);
		}
	}
	if (%safety >= 10000)
	{
		announce("Safety limit reached");
	}
}

function createTrackNodeShapes(%brick)
{
    %center = vectorAdd(%brick.getPosition(), "0 0 0.3");

    %dir = %brick.dynamicDir;

    %startingPos = vectorAdd(%brick.Point[getWord(%dir, 0)], "0 0 0.3");
    createBoxAt(%startingPos, "1 0 0 1", 0.1);

    //get curves
    for (%i = 1; %i < getWordCount(%dir); %i++)
    {
        %currDir = getWord(%dir, %i);
        %pos = vectorAdd(%brick.Point[%currDir], "0 0 0.3");
        createBoxAt(%pos, "1 0 0 1", 0.1);

        %connectionCenter = vectorScale(vectorAdd(vectorAdd(vectorScale(%center, 2), %pos), %startingPos), 1.0/4.0);
        %preCenter = vectorScale(vectorAdd(%center, %startingPos), 0.5);
        %postCenter = vectorScale(vectorAdd(%center, %pos), 0.5);
        %connectionpre = vectorScale(vectorAdd(vectorAdd(vectorScale(%preCenter, 2), %connectionCenter), %startingPos), 1.0/4.0);
        %connectionpost = vectorScale(vectorAdd(vectorAdd(vectorScale(%postCenter, 2), %connectionCenter), %pos), 1.0/4.0);

        %preprecenter = vectorScale(vectorAdd(%center, %startingPos), 0.5);
        drawLine(%pos, %connectionCenter, "1 1 0 0.2", 0.05);
        drawLine(%startingPos, %connectionCenter, "1 1 0 0.2", 0.05);

        drawLine(%startingPos, %connectionPre, "1 0 1 0.2", 0.05).setShapeName(%count++ - 1);
        drawLine(%connectionPre, %connectionCenter, "1 0 1 0.2", 0.05).setShapeName(%count++ - 1);
        drawLine(%connectionPost, %connectionCenter, "1 0 1 0.2", 0.05).setShapeName(%count++ - 1);
        drawLine(%pos, %connectionPost, "1 0 1 0.2", 0.05).setShapeName(%count++ - 1);

        createBoxAt(%connectionCenter, "1 0 0 1", 0.1);
    }
}

function followVehicle(%obj, %veh)
{
	cancel(%obj.followSchedule);

	if ((%frontDist = %obj.getDatablock().frontDist) !$= "")
	{
		%trainFront = vectorAdd(%obj.getPosition(), vectorScale(%obj.getForwardVector(), %frontDist));
		if (isObject(%obj.shapeBox))
			%obj.shapeBox.delete();
	} 
	else
	{
		return;
	}

	if ((%backDist = %veh.getDatablock().backDist) !$= "")
	{
		%trainBack = vectorSub(%veh.getPosition(), vectorScale(%veh.getForwardVector(), %backDist));
		if (isObject(%veh.shapeBox))
			%veh.shapeBox.delete();
	}
	else
	{
		return;
	}

	//look at veh
	%normal = vectorNormalize(vectorSub(%veh.getPosition(), %obj.getPosition()));
	%directionNormal = vectorNormalize(vectorScale(%normal, vectorDot(%obj.getForwardVector(), %normal)));
	%dir = vectorDot(%obj.getForwardVector(), %normal);
	%vector = vectorNormalize(vectorCross(vectorNormalize(%directionNormal), "0 0 1"));

	if (getSubStr(%vector, 1, strLen(%vector)) $= "1 0 0")
	{
		if (%vector $= "-1 0 0")
		{
			%xyz = "0 0 1";
			%u = 3.14159;
		}
		else
		{
			%xyz = "0 0 0";
			%u = "0";
		}
	}
	else
	{
		if (%dir < 0)
		{
			%xyz = vectorNormalize(vectorCross("-1 0 0", %vector)); //rotation axis
			%u = mACos(vectorDot("1 0 0", %vector)) * 1; //rotation value
		}
		else
		{
			%xyz = vectorNormalize(vectorCross("1 0 0", %vector)); //rotation axis
			%u = mACos(vectorDot("1 0 0", %vector)) * -1; //rotation value
		}
	}

	// %newRot = %xyz SPC %u;

	//adjust position to be right distance from veh
	%offset = vectorSub(%trainBack, %trainFront);
	%newPos = vectorAdd(%obj.getPosition(), %offset);

	%obj.setTransform(%newPos SPC %newRot);

	//adopt veh velocity, applied onto current forwardVec
	%newVel = vectorScale(%obj.getForwardVector(), vectorLen(%veh.getVelocity()) / 2);
	%obj.setVelocity(%newVel);

	%obj.followSchedule = schedule(10, %obj, followVehicle, %obj, %veh);
}

function trainLockLoop(%train)
{
	if (!isObject(%train) || !isFunction(%train.getClassName(), "getForwardVector"))
	{
		return;
	}

	%transform = %train.getTransform();
	%trainPos = getWords(%transform, 0, 2);
	%trainRot = getWords(%transform, 3, 6);
	%trainVel = %train.getVelocity();

	%nextTrainPos = vectorAdd(%trainPos, vectorScale(%trainVel, 1/33));



	//check for track brick underneath
	%start = vectorAdd(%trainPos, "0 0 0.1");
	%end = "0 0" SPC getWord(%train.getWorldBox(), 2) - getWord(%start, 2) - 2;

	%ray = containerRaycast(%start, vectorAdd(%start, %end), $TypeMasks::fxBrickObjectType);
	%search = getWords(%trainPos, 0, 1) SPC getWord(%train.getObjectBox(), 2) + getWord(%trainPos, 2);

	initContainerBoxSearch(%search, "1.5 1.5 0.8", $TypeMasks::fxBrickObjectType);
	%shortestDist = 1000;
	while (isObject(%next = containerSearchNext()))
	{
		if (%next.getDatablock().isTrainTrack && vectorDist(%next.getPosition(), %trainPos) < %shortestDist)
		{
			%hit = %next;
			%shortestDist = vectorDist(%next.getPosition(), %trainPos);
		}
	}

	if (!isObject(%hit)) //redo the search with a position decided by raycast
	{
		%ray = containerRaycast(%start, vectorAdd(%start, %end), $TypeMasks::fxBrickObjectType);
		%search = getWords(%ray, 1, 3);
		
		initContainerBoxSearch(%search, "1.5 1.5 0.8", $TypeMasks::fxBrickObjectType);
		%shortestDist = 1000;
		while (isObject(%next = containerSearchNext()))
		{
			if (%next.getDatablock().isTrainTrack && vectorDist(%next.getPosition(), %trainPos) < %shortestDist)
			{
				%hit = %next;
				%shortestDist = vectorDist(%next.getPosition(), %trainPos);
			}
		}
	}

	if (!isObject(%hit) || !%hit.getDatablock().isTrainTrack || %hit.dynamicDir $= "") //no brick found, premature exit
	{
		cancel(%train.trainLoopSchedule);
		%train.trainLoopSchedule = schedule(200, %train, trainLockLoop, %train);
		%train.lastBrick = "";
		return;
	}



	//get track brick data
	%brick = %hit;
	%brickDB = %hit.getDatablock();
	%center = vectorAdd(%brick.getPosition(), "0 0 0.3");

	%dir = %brick.dynamicDir;
	%switchMode = isObject(%brickDB.switchBrick0);
	%blocked = %brickDB.blocked;
	%crossing = %brickDB.isCrossing;
	if (%switchMode)
	{
		//brick will already be appropriate datablock with relevant values
		%switchEnabled = 1;
	}

	%startingPos = vectorAdd(%brick.Point[getWord(%dir, 0)], "0 0 0.3");



    //get curves and unblocked curves (if part of a switch)
    //different code if its a crossing brick
    if (!%crossing) //normal or switch brick
    {
        // if (%brick.curve[0] $= "")
        // {
            for (%i = 1; %i < getWordCount(%dir); %i++)
            {
                %currDir = getWord(%dir, %i);
                %pos = vectorAdd(%brick.Point[%currDir], "0 0 0.3");

                %connectionCenter = vectorScale(vectorAdd(vectorAdd(vectorScale(%center, 2), %pos), %startingPos), 1.0/4.0);
                %preCenter = vectorScale(vectorAdd(%center, %startingPos), 0.5);
                %postCenter = vectorScale(vectorAdd(%center, %pos), 0.5);
                %connectionpre = vectorScale(vectorAdd(vectorAdd(vectorScale(%preCenter, 2), %connectionCenter), %startingPos), 1.0/4.0);
                %connectionpost = vectorScale(vectorAdd(vectorAdd(vectorScale(%postCenter, 2), %connectionCenter), %pos), 1.0/4.0);

                if (strPos(" " @ %blocked @ " ", " " @ %i @ " ") < 0 && %switchEnabled) //this is a switch, only allow unblocked routes
                {
                    %curveUnblocked[%totalBlockedCurves++ - 1] = %startingPos TAB %connectionCenter;
                    %curveUnblocked[%totalBlockedCurves++ - 1] = %connectionCenter TAB %pos;
                }

                %curve[%totalCurves++ - 1] = %startingPos TAB %connectionPre;
                %curve[%totalCurves++ - 1] = %connectionPre TAB %connectionCenter;
                %curve[%totalCurves++ - 1] = %connectionCenter TAB %connectionPost;
                %curve[%totalCurves++ - 1] = %connectionPost TAB %pos;

                %curveCountPerTurn = 4;
            }

            for (%i = 0; %i < %totalCurves; %i++)
            {
                %brick.curve[%i] = %curve[%i];
            }
            %brick.totalCurves = %totalCurves;
        // }
        // else
        // {
        //     for (%i = 0; %i < %brick.totalCurves; %i++)
        //     {
        //         %curve[%i] = %brick.curve[%i];
        //     }
        //     %totalCurves = %brick.totalCurves;
        // }
    }
    else //its a crossing brick
    {
        for (%i = 0; %i < getWordCount(%dir); %i = %i + 2) //separate the connections into pairs
        {
            %firstDir = getWord(%dir, %i);
            %secondDir = getWord(%dir, %i + 1);
            %pos1 = vectorAdd(%brick.Point[%firstDir], "0 0 0.3");
            %pos2 = vectorAdd(%brick.Point[%secondDir], "0 0 0.3");

            %curve[%totalCurves++ - 1] = %pos1 TAB %pos2;
        }

        //force non-switch state since track has no front side
        %switchEnabled = 0;
    }

    if (%hit != %train.lastBrick)
    {       
        //call exit and enter events
        %train.lastBrick.onTrainLeaveTrack(%train);
        %hit.onTrainEnterTrack(%train);
    }

    if (%hit != %train.lastBrick && isObject(%train.lastBrick) && getWordCount(%dir) > 2) //we're entering a new track, check if right side
    {
        //get position of last brick relative to this one
        %lastxy = getWords(%train.lastBrick.getPosition(), 0, 1);
        %hitxy = getWords(%hit.getPosition(), 0, 1);

        %brickDir = vectorSub(%lastxy, %hitxy);

        %brickDirX = mFloor(getWord(%brickDir, 0) + 0.5);
        %brickDirY = mFloor(getWord(%brickDir, 1) + 0.5);

        if (%brickDirX == 0) //must be north/south
        {
            if (%brickDirY < 0) //south
                %brickCompassDir = "S";
            else if (%brickDirY > 0) //north
                %brickCompassDir = "N";
            else //same location???
                %brickCompassDir = "ERROR";
        }
        else if (%brickDirX < 0) //west
        {
            if (%brickDirY < 0) //south
                %brickCompassDir = "SW";
            else if (%brickDirY > 0) //north
                %brickCompassDir = "NW";
            else
                %brickCompassDir = "W";
        }
        else if (%brickDirX > 0) //east
        {
            if (%brickDirY < 0) //south
                %brickCompassDir = "SE";
            else if (%brickDirY > 0) //north
                %brickCompassDir = "NE";
            else
                %brickCompassDir = "E";
        }

        if (getWord(%dir, 0) $= %brickCompassDir && !%crossing)
        {
            if (!%switchEnabled) //we have entered the correct side + not switch, pick correct direction based on dir mode
            {
                %newPos = "";
                %bestCurve = "";

                if (%train.snapMode $= "Left") //always take leftmost one
                {
                    %bestCurve = %curve[0];
                }
                else if (%train.snapMode $= "Right") //always take rightmost one
                {
                    %bestCurve = %curve[%totalCurves - %curveCountPerTurn];
                }
                else //take center/left depending whats available
                {
                    %clockwise = "N NE E SE S SW W NW N NE E SE S SW W NW";
                    
                    for (%j = 0; %j < getWordCount(%clockwise); %j++)
                    {
                        %search = getWord(%clockwise, %j);
                        if (%search $= %brickCompassDir)
                        {
                            break;
                        }
                    }

                    //rotate current direction based on angleID (90 deg)
                    %j = %j + 4;
                    %opposite = getWord(%clockwise, %j);

                    //check for opposite, take it if its available. otherwise take leftmost (0)
                    for (%j = 0; %j < getWordCount(%dir); %j++)
                    {
                        %search = getWord(%dir, %j);
                        if (%search $= %opposite)
                        {
                            %found = 1;
                            break;
                        }
                    }

                    if (%found)
                        %bestCurve = %curve[(%j-1) * %curveCountPerTurn];
                    else
                        %bestCurve = %curve[0];
                }

                //get the newPos based on the bestCurve
                %newPos = getWords(absDistanceFromVector(getField(%bestCurve, 0), getField(%bestCurve, 1), %trainPos), 2, 4);
                
                %usingSmartSnap = 1;
            }
            else if (%switchEnabled) //snap only to switch enabled paths
            {
                %newPos = "";
                %bestCurve = "";
                %bestDist = 10000;
                for (%i = 0; %i < %totalBlockedCurves; %i++)
                {
                    %currCurve = %curveUnblocked[%i];
                    %ret = absDistanceFromVector(getField(%currCurve, 0), getField(%currCurve, 1), %trainPos);
    
                    %currDist = getWord(%ret, 0);
    
                    %currPos = getWords(%ret, 2, 4); //position on curve closest to %trainPos

                    if (%currDist < %bestDist)
                    {
                        %newPos = %currPos;
                        %bestCurve = %currCurve;
                        %bestDist = %currDist;
                    }
                }

                %usingSmartSnap = 1;
            }

            // if (isObject(%train.client) || isObject(%train.getControllingClient()))
            // { 
            //  %client = isObject(%train.client) ? %train.client : %train.getControllingClient();
            //  %client.centerPrint("Switch?: " @ %switchEnabled SPC "| blocked: " @ %blocked SPC "| totalBlockedCurves: " @ %totalBlockedCurves, -1);
            // }
        }
    }
    
    if (!%usingSmartSnap) //default behavior
    {
        //shift position to nearest curve
        %newPos = "";
        %bestCurve = "";
        %bestDist = 10000;
        for (%i = 0; %i < %totalCurves; %i++)
        {
            %currCurve = %curve[%i];
            %ret = absDistanceFromVector(getField(%currCurve, 0), getField(%currCurve, 1), %trainPos);
            %currDist = getWord(%ret, 0);
            %currPos = getWords(%ret, 2, 4); //position on curve closest to %trainPos

            if (%currDist > 100 && %crossing) //just derailed off a crossing brick, premature end or else we'll snap to the other direction
            {
                cancel(%train.trainLoopSchedule);
                %train.trainLoopSchedule = schedule(200, %train, trainLockLoop, %train);
                %train.lastBrick = "";
                return;
            }

            if (%currDist < %bestDist)
            {
                %newPos = %currPos;
                %bestCurve = %currCurve;
                %bestDist = %currDist;
            }
        }
    }

    //adjust current vehicle velocity to align with track
    %normal = vectorNormalize(vectorSub(getField(%bestCurve, 1), getField(%bestCurve, 0)));
    %newVel = vectorScale(%normal, vectorDot(%normal, %trainVel));
    //calculate new rotation of train
    %directionNormal = vectorNormalize(vectorScale(%normal, vectorDot(%train.getForwardVector(), %normal)));
    %dir = vectorDot(%train.getForwardVector(), %normal);
    %vector = vectorNormalize(vectorCross(vectorNormalize(%directionNormal), "0 0 1"));

    if (getSubStr(%vector, 1, strLen(%vector)) $= "1 0 0") //manual fix for facing along x axis
    {
        if (%vector $= "-1 0 0")
        {
            %xyz = "0 0 1";
            %u = 3.14159;
        }
        else
        {
            %xyz = "0 0 0";
            %u = "0";
        }
    }
    else
    {
        if (%dir < 0)
        {
            %xyz = vectorNormalize(vectorCross("-1 0 0", %vector)); //rotation axis
            %u = mACos(vectorDot("1 0 0", %vector)) * 1; //rotation value
        }
        else
        {
            %xyz = vectorNormalize(vectorCross("1 0 0", %vector)); //rotation axis
            %u = mACos(vectorDot("1 0 0", %vector)) * -1; //rotation value
        }
    }

    //reduce train jitter on ramps
    %newRot = %xyz SPC %u;
    %angle = mACos(vectorDot(vectorNormalize(getWords(%trainVel, 0, 1)), %normal));
    if (%angle < 0.1 || %angle > 3.04159 || vectorLen(%trainVel) < 1.5)
    {
        %newRot = getWords(%train.getTransform(), 3, 6);
    }
    // %newRot = %trainRot;

    //assign all values to train
    %newTransform = getWords(%newPos, 0, 1) SPC getWord(%trainPos, 2) SPC %newRot;

    %train.setTransform(%newTransform);
    %train.setVelocity(getWords(%newVel, 0, 1) SPC getWord(%trainVel, 2)/4*3);

    //schedule next loop
    cancel(%train.trainLoopSchedule);
    %train.trainLoopSchedule = schedule(33, %train, trainLockLoop, %train);
    %train.lastBrick = %hit;

    if (isObject(%train.client) || isObject(%train.getControllingClient()))
    { 
        if (!isObject(%train.client))
            centerprintControls(%train.getControllingClient(), %train);
        else
            centerprintControls(%train.client, %train);
    }

    return;
}

//////
//ui//
//////

function centerprintControls(%client, %train)
{
	if (%train.snapMode $= "")
		%train.snapMode = "STRAIGHT";

	%control = strUpr(%train.snapMode);

	%str1 = "<br><br><br><br><br><br><color:ffffff>[" @ %control @ "] <br>";
	%str2 = "<color:ffff00>Light<color:ffffff>: Cycle Direction          <color:ffff00>Brick Controls<color:ffffff>: Set Direction";
	%client.centerprint(%str1 SPC %str2, 1);
}

package SupportTrainControls
{
	function serverCmdLight(%cl)
	{
		if (isObject(%pl = %cl.player) && isObject(%veh = %cl.player.getObjectMount()) && isEventPending(%veh.trainLoopSchedule)
			&& %veh.getControllingClient() == %cl)
		{
			if (%veh.snapMode $= "STRAIGHT")
				%veh.snapMode = "RIGHT";
			else if (%veh.snapMode $= "RIGHT")
				%veh.snapMode = "LEFT";
			else
				%veh.snapMode = "STRAIGHT";

			centerprintControls(%cl, %veh);
			return;
		}

		return parent::serverCmdLight(%cl);
	}

	function serverCmdShiftBrick(%cl, %y, %x, %z)
	{
		if (isObject(%pl = %cl.player) && isObject(%veh = %cl.player.getObjectMount()) && isEventPending(%veh.trainLoopSchedule)
			&& %veh.getControllingClient() == %cl)
		{
			// %compassVec = getCompassVec(%cl);
			// %temp = %x;

			// switch$ (%compassVec) {
			// 	case "1 0 0":   %x = %x;        %y = %y;
			// 	case "0 1 0":   %x = -%y;       %y = %temp;
			// 	case "-1 0 0":  %x = -%x;       %y = -%y;
			// 	case "0 -1 0":  %x = %y;        %y = -%temp;
			// 	default:        %x = %x;        %y = %y;
			// }

			if (%x > 0)
				%veh.snapMode = "LEFT";
			else if (%x < 0)
				%veh.snapMode = "RIGHT";
			else if (%y > 0)
				%veh.snapMode = "STRAIGHT";

			centerprintControls(%cl, %veh);
			return;
		}

		return parent::serverCmdShiftBrick(%cl, %y, %x, %z);
	}

	function serverCmdSuperShiftBrick(%cl, %y, %x, %z)
	{
		if (isObject(%pl = %cl.player) && isObject(%veh = %cl.player.getObjectMount()) && isEventPending(%veh.trainLoopSchedule)
			&& %veh.getControllingClient() == %cl)
		{
			// %compassVec = getCompassVec(%cl);
			// %temp = %x;

			// switch$ (%compassVec) {
			// 	case "1 0 0":   %x = %x;        %y = %y;
			// 	case "0 1 0":   %x = -%y;       %y = %temp;
			// 	case "-1 0 0":  %x = -%x;       %y = -%y;
			// 	case "0 -1 0":  %x = %y;        %y = -%temp;
			// 	default:        %x = %x;        %y = %y;
			// }

			if (%x > 0)
				%veh.snapMode = "LEFT";
			else if (%x < 0)
				%veh.snapMode = "RIGHT";
			else if (%y > 0)
				%veh.snapMode = "STRAIGHT";

			centerprintControls(%cl, %veh);
			return;
		}

		return parent::serverCmdSuperShiftBrick(%cl, %y, %x, %z);
	}
};
activatePackage(SupportTrainControls);